"""This module provides models for calculating various body parameters including surface area,
weight ratio, basal blood flow ratio, thermal conductance, and thermal capacity.

The values of a NumPy array containing 17 elements correspond to the body parts defined in JOS3BodyParts.
"""

import numpy as np

from pythermalcomfort.classes_return import JOS3BodyParts
from pythermalcomfort.jos3_functions.matrix import IDICT, NUM_NODES
from pythermalcomfort.jos3_functions.parameters import Default
from pythermalcomfort.utilities import body_surface_area


def validate_body_parameters(
    height: float,
    weight: float,
    age: int,
    body_fat: float,
):
    """Validate the body parameters: height, weight, age, and body fat percentage.

    Parameters
    ----------
    height : float, optional
        The height of the person in meters.
    weight : float, optional
        The weight of the person in kilograms.
    age : int, optional
        The age of the person in years.
    body_fat : float, optional
        The body fat percentage as a fraction of total body mass.

    Raises
    ------
    ValueError
        If any of the parameters are out of the specified range.

    Examples
    --------
    Validate parameters for a person with height 1.80 meters, weight 75 kilograms, age 30 years, and body fat 20%:

    .. code-block:: python

        validate_body_parameters(height=1.80, weight=75, age=30, body_fat=20)

    """
    if not (0.5 <= height <= 3.0):
        raise ValueError("Height must be in the range [0.5, 3.0] meters.")

    if not (20.0 <= weight <= 200.0):
        raise ValueError("Weight must be in the range [20.0, 200.0] kilograms.")

    if not (5 <= age <= 100):
        raise ValueError("Age must be in the range [5, 100] years.")

    if not (1 <= body_fat <= 90):
        raise ValueError(
            "Body fat percentage must be in the range [1, 90] (1% to 90%).",
        )


def to_array_body_parts(inp) -> np.ndarray:
    """Create a NumPy array of shape (17,) with the given input.

    Parameters
    ----------
    inp : int, float, dict, list, ndarray
        The value(s) to use when creating the 17-element array.
        Supports int, float, dict (with BODY_NAMES as keys), list, and ndarray types.

    Returns
    -------
    ndarray
        A NumPy array of shape (17,) with the specified values.

    Raises
    ------
    ValueError
        If the input type is not supported or if the input list or ndarray is not of length 17.

    """
    if isinstance(inp, int | float):
        return np.full(Default.num_body_parts, inp)
    if isinstance(inp, dict):
        return np.array([inp[key] for key in JOS3BodyParts.get_attribute_names()])
    if isinstance(inp, list | np.ndarray):
        inp = np.asarray(inp)
        if inp.shape == (Default.num_body_parts,):
            return inp.copy()
        error_msg = (
            f"The input list or ndarray is not of length {Default.num_body_parts}"
        )
        raise ValueError(error_msg)
    error_msg = f"Unsupported input type: {type(inp)}. Supported types: int, float, list, dict, ndarray"
    raise ValueError(error_msg)


def bsa_rate(
    height: float,
    weight: float,
    bsa_equation: str,
) -> float | np.ndarray:
    """Calculate the ratio of body surface area (BSA) to the standard body.

    This function computes the ratio of an individual's body surface area to that of a standard body,
    based on the given height, weight, and BSA calculation equation.

    Parameters
    ----------
    height : float, optional
        The height of the person in meters.
    weight : float, optional
        The weight of the person in kilograms.
    bsa_equation : str, optional
        The equation name for BSA calculation. Choose one from BodySurfaceAreaEquations

    Returns
    -------
    float, np.ndarray
        The ratio of the individual's BSA to the standard body BSA.

    Raises
    ------
    ValueError
        If the parameters are out of the specified range.

    Examples
    --------
    Calculate the BSA ratio for a person with height 1.80 meters and weight 75 kilograms:

    .. code-block:: python

        bsa_ratio = bsa_rate(height=1.80, weight=75, bsa_equation="dubois")
        print(bsa_ratio)

    """
    bsa_all = body_surface_area(
        height=height,
        weight=weight,
        formula=bsa_equation,
    )
    return (
        bsa_all / Default.local_bsa.sum()
    )  # The bsa ratio to the standard body (1.87m2)


def local_bsa(
    height: float,
    weight: float,
    bsa_equation: str,
) -> np.ndarray:
    """Calculate local body surface area (BSA) in square meters.

    The local body surface area has been derived from 65MN.
    The head has been divided into head and neck based on Smith's model.
        head = 0.1396 * 0.1117 / 0.1414 (65MN_Head * Smith_Head / Smith_Head+neck)
        neck = 0.1396 * 0.0297 / 0.1414 (65MN_Head * Smith_Neck / Smith_Head+neck)

    Parameters
    ----------
    height : float, optional
        Body height in meters.
    weight : float, optional
        Body weight in kilograms.
    bsa_equation : str, optional
        The equation name for BSA calculation. Choose one from BodySurfaceAreaEquations.

    Returns
    -------
    np.ndarray
        Local body surface area (BSA) in square meters.

    Raises
    ------
    ValueError
        If any of the parameters are out of the specified range.

    Examples
    --------
    Calculate the local BSA for a person with height 1.80 meters and weight 75 kilograms:

    .. code-block:: python

        local_bsa_values = local_bsa(height=1.80, weight=75, bsa_equation="dubois")
        print(local_bsa_values)

    """
    bsa_ratio = bsa_rate(height=height, weight=weight, bsa_equation=bsa_equation)
    return Default.local_bsa * bsa_ratio


def weight_rate(weight: float) -> float:
    """Calculate the ratio of the body weight to the standard body.

    Parameters
    ----------
    weight : float, optional
        The body weight in kilograms.

    Returns
    -------
    float
        The ratio of the body weight to the standard body.

    """
    return weight / Default.weight


def bfb_rate(
    height: float,
    weight: float,
    bsa_equation: str,
    age: int,
    ci: float,
) -> float:
    """Calculate the ratio of basal blood flow (BFB) to the standard body.

    This function computes the basal blood flow rate based on the given height, weight,
    BSA calculation equation, age, and cardiac index.

    Parameters
    ----------
    height : float, optional
        Body height in meters.
    weight : float, optional
        Body weight in kilograms.
    bsa_equation : str, optional
        The equation name for BSA calculation. Choose one from BodySurfaceAreaEquations.
    age : int, optional
        Age in years.
    ci : float, optional
        Cardiac index in L/min/m².

    Returns
    -------
    float
        Basal blood flow rate.

    Raises
    ------
    ValueError
        If any of the parameters are out of the specified range.

    Examples
    --------
    Calculate the BFB rate for a person with height 1.80 meters, weight 75 kilograms, age 30 years, and cardiac index 2.59:

    .. code-block:: python

        bfb_ratio = bfb_rate(height=1.80, weight=75, age=30, ci=2.59)
        print(bfb_ratio)

    """
    ci *= 60  # Convert unit from L/min/m² to L/h/m²

    # Adjust cardiac index based on age
    if age < 50:
        ci *= 1
    elif age < 60:
        ci *= 0.85
    elif age < 70:
        ci *= 0.75
    else:  # age >= 70
        ci *= 0.7

    bsa_ratio = bsa_rate(height, weight, bsa_equation)
    bfb_all = ci * bsa_ratio * Default.local_bsa.sum()  # Total BFB in L/h
    return bfb_all / Default.blood_flow_rate  # Ratio to the standard body (290 L/h)


def conductance(
    height: float,
    weight: float,
    bsa_equation: str,
    fat: float,
):
    """Calculate thermal conductance between layers [W/K].

    Parameters
    ----------
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    fat : float, optional
        Body fat rate [%].

    Returns
    -------
    conductance : numpy.ndarray
        Thermal conductance between layers [W/K].
        The shape is (NUM_NODES, NUM_NODES).

    """
    if fat < 12.5:
        cdt_cr_sk = np.array(
            [
                1.341,
                0.930,
                1.879,
                1.729,
                2.370,
                1.557,
                1.018,
                2.210,
                1.557,
                1.018,
                2.210,
                2.565,
                1.378,
                3.404,
                2.565,
                1.378,
                3.404,
            ],
        )
    elif fat < 17.5:
        cdt_cr_sk = np.array(
            [
                1.311,
                0.909,
                1.785,
                1.643,
                2.251,
                1.501,
                0.982,
                2.183,
                1.501,
                0.982,
                2.183,
                2.468,
                1.326,
                3.370,
                2.468,
                1.326,
                3.370,
            ],
        )
    elif fat < 22.5:
        cdt_cr_sk = np.array(
            [
                1.282,
                0.889,
                1.698,
                1.563,
                2.142,
                1.448,
                0.947,
                2.156,
                1.448,
                0.947,
                2.156,
                2.375,
                1.276,
                3.337,
                2.375,
                1.276,
                3.337,
            ],
        )
    elif fat < 27.5:
        cdt_cr_sk = np.array(
            [
                1.255,
                0.870,
                1.618,
                1.488,
                2.040,
                1.396,
                0.913,
                2.130,
                1.396,
                0.913,
                2.130,
                2.285,
                1.227,
                3.304,
                2.285,
                1.227,
                3.304,
            ],
        )
    else:  # fat >= 27.5
        cdt_cr_sk = np.array(
            [
                1.227,
                0.852,
                1.542,
                1.419,
                1.945,
                1.346,
                0.880,
                1.945,
                1.346,
                0.880,
                1.945,
                2.198,
                1.181,
                3.271,
                2.198,
                1.181,
                3.271,
            ],
        )

    cdt_cr_ms = np.zeros(Default.num_body_parts)  # core to muscle [W/K]
    cdt_ms_fat = np.zeros(Default.num_body_parts)  # muscle to fat [W/K]
    cdt_fat_sk = np.zeros(Default.num_body_parts)  # fat to skin [W/K]

    # head and pelvis consists of 65MN's conductances
    cdt_cr_ms[0] = 1.601  # head
    cdt_ms_fat[0] = 13.222
    cdt_fat_sk[0] = 16.008
    cdt_cr_ms[4] = 3.0813  # pelvis
    cdt_ms_fat[4] = 10.3738
    cdt_fat_sk[4] = 41.4954

    # vessel to core
    # The shape is a cylinder.
    # It is assumed that the inner is vascular radius, 2.5mm and the outer is
    # stolwijk's core radius.
    # The heat transfer coefficient of the core is assumed as the Michel's
    # counter-flow model 0.66816 [W/(m･K)].
    cdt_ves_cr = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            0.586,
            0.383,
            1.534,
            0.586,
            0.383,
            1.534,
            0.810,
            0.435,
            1.816,
            0.810,
            0.435,
            1.816,
        ],
    )
    # superficial vein to skin
    cdt_sfv_sk = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            57.735,
            37.768,
            16.634,
            57.735,
            37.768,
            16.634,
            102.012,
            54.784,
            24.277,
            102.012,
            54.784,
            24.277,
        ],
    )

    # art to vein (counter-flow) [W/K]
    # The data has been derived Mitchell's model.
    # The values = 15.869 [W/(m･K)] * the segment length [m]
    cdt_art_vein = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            0.537,
            0.351,
            0.762,
            0.537,
            0.351,
            0.762,
            0.826,
            0.444,
            0.992,
            0.826,
            0.444,
            0.992,
        ],
    )

    # Changes values by body size based on the standard body.
    wr = weight_rate(weight)
    bsar = bsa_rate(height, weight, bsa_equation)
    # head, neck (Sphere shape)
    # TODO we are multiplying zeros by a value
    cdt_cr_sk[:2] *= wr / bsar
    cdt_cr_ms[:2] *= wr / bsar
    cdt_ms_fat[:2] *= wr / bsar
    cdt_fat_sk[:2] *= wr / bsar
    cdt_ves_cr[:2] *= wr / bsar
    cdt_sfv_sk[:2] *= wr / bsar
    cdt_art_vein[:2] *= wr / bsar
    # Others (Cylinder shape)
    cdt_cr_sk[2:] *= bsar**2 / wr
    cdt_cr_ms[2:] *= bsar**2 / wr
    cdt_ms_fat[2:] *= bsar**2 / wr
    cdt_fat_sk[2:] *= bsar**2 / wr
    cdt_ves_cr[2:] *= bsar**2 / wr
    cdt_sfv_sk[2:] *= bsar**2 / wr
    cdt_art_vein[2:] *= bsar**2 / wr

    cdt_whole = np.zeros((NUM_NODES, NUM_NODES))
    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        # Dictionary of indices in each body segment
        # key = layer name, value = index of matrix
        index_of = IDICT[bn]

        # Common
        cdt_whole[index_of["artery"], index_of["vein"]] = cdt_art_vein[i]  # art to vein
        cdt_whole[index_of["artery"], index_of["core"]] = cdt_ves_cr[i]  # art to cr
        cdt_whole[index_of["vein"], index_of["core"]] = cdt_ves_cr[i]  # vein to cr

        # Only limbs
        if i >= 5:
            cdt_whole[index_of["sfvein"], index_of["skin"]] = cdt_sfv_sk[i]  # sfv to sk

        # If the segment has a muscle or fat layer
        if index_of["muscle"] is not None:  # or not indexof["fat"] is None
            cdt_whole[index_of["core"], index_of["muscle"]] = cdt_cr_ms[i]  # cr to ms
            cdt_whole[index_of["muscle"], index_of["fat"]] = cdt_ms_fat[i]  # ms to fat
            cdt_whole[index_of["fat"], index_of["skin"]] = cdt_fat_sk[i]  # fat to sk

        else:
            cdt_whole[index_of["core"], index_of["skin"]] = cdt_cr_sk[i]  # cr to sk

    # Creates a symmetrical matrix
    cdt_whole = cdt_whole + cdt_whole.T

    return cdt_whole.copy()


def capacity(
    height: float,
    weight: float,
    bsa_equation: str,
    age: int,
    ci: float,
) -> np.ndarray:
    """Calculate the thermal capacity [J/K].

    The values of vascular and central blood capacity have been derived from
    Yokoyama's model.
    The specific heat of blood is assumed as 1.0 [kcal/L.K].

    Parameters
    ----------
    height : float, optional
        Body height [m].
    weight : float, optional
        Body weight [kg].
    bsa_equation : str, optional
        The equation name (str) of bsa calculation. Choose one from BodySurfaceAreaEquations.
    age : int, optional
        Age [years].
    ci : float, optional
        Cardiac index [L/min/㎡].

    Returns
    -------
    np.ndarray
        Thermal capacity [J/K].
        The shape is (NUM_NODES).

    """
    # Define capacities [Wh/K]

    # artery [Wh/K]
    cap_art = np.array(
        [
            0.096,
            0.025,
            0.12,
            0.111,
            0.265,
            0.0186,
            0.0091,
            0.0044,
            0.0186,
            0.0091,
            0.0044,
            0.0813,
            0.04,
            0.0103,
            0.0813,
            0.04,
            0.0103,
        ],
    )

    # vein [Wh/K]
    cap_vein = np.array(
        [
            0.321,
            0.085,
            0.424,
            0.39,
            0.832,
            0.046,
            0.024,
            0.01,
            0.046,
            0.024,
            0.01,
            0.207,
            0.1,
            0.024,
            0.207,
            0.1,
            0.024,
        ],
    )

    # superficial vein [Wh/K]
    cap_sfv = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            0.025,
            0.015,
            0.011,
            0.025,
            0.015,
            0.011,
            0.074,
            0.05,
            0.021,
            0.074,
            0.05,
            0.021,
        ],
    )

    # central blood [Wh/K]
    cap_cb = 1.999

    # core [Wh/K]
    cap_cr = np.array(
        [
            1.7229,
            0.564,
            10.2975,
            9.3935,
            4.488,
            1.6994,
            1.1209,
            0.1536,
            1.6994,
            1.1209,
            0.1536,
            5.3117,
            2.867,
            0.2097,
            5.3117,
            2.867,
            0.2097,
        ],
    )

    # muscle [Wh/K]
    cap_ms = np.array(
        [
            0.305,
            0.0,
            0.0,
            0.0,
            7.409,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )

    # fat [Wh/K]
    cap_fat = np.array(
        [
            0.203,
            0.0,
            0.0,
            0.0,
            1.947,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
    )

    # skin [Wh/K]
    cap_sk = np.array(
        [
            0.1885,
            0.058,
            0.441,
            0.406,
            0.556,
            0.126,
            0.084,
            0.088,
            0.126,
            0.084,
            0.088,
            0.334,
            0.169,
            0.107,
            0.334,
            0.169,
            0.107,
        ],
    )

    # Adjust capacities based on body parameters
    bfbr = bfb_rate(height, weight, bsa_equation, age, ci)
    wr = weight_rate(weight)
    cap_art *= bfbr
    cap_vein *= bfbr
    cap_sfv *= bfbr
    cap_cb *= bfbr
    cap_cr *= wr
    cap_ms *= wr
    cap_fat *= wr
    cap_sk *= wr

    # Initialize capacity array
    cap_whole = np.zeros(NUM_NODES)
    cap_whole[0] = cap_cb

    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        # Dictionary of indices in each body segment key = layer name, value = index of matrix
        index_of = IDICT[bn]

        # Common
        cap_whole[index_of["artery"]] = cap_art[i]
        cap_whole[index_of["vein"]] = cap_vein[i]
        cap_whole[index_of["core"]] = cap_cr[i]
        cap_whole[index_of["skin"]] = cap_sk[i]

        # Only limbs
        if i >= 5:
            cap_whole[index_of["sfvein"]] = cap_sfv[i]

        # If the segment has a muscle or fat layer
        if index_of["muscle"] is not None:  # or not indexof["fat"] is None
            cap_whole[index_of["muscle"]] = cap_ms[i]
            cap_whole[index_of["fat"]] = cap_fat[i]

    cap_whole *= 3600  # Convert [Wh/K] to [J/K]
    return cap_whole


def pass_values_to_jos3_body_parts(values, round_digits=2, body_parts=None):
    if body_parts is None:
        body_parts = JOS3BodyParts.get_attribute_names()
    return JOS3BodyParts(
        **{
            name: value
            for name, value in zip(
                body_parts, np.round(values, round_digits), strict=False
            )
        },
    )



import numpy as np
import pytest

from pythermalcomfort.utilities import (
    Units,
    body_surface_area,
    clo_area_factor,
    clo_correction_factor_environment,
    clo_dynamic_ashrae,
    clo_dynamic_iso,
    clo_insulation_air_layer,
    clo_intrinsic_insulation_ensemble,
    clo_total_insulation,
    f_svv,
    running_mean_outdoor_temperature,
    transpose_sharp_altitude,
    units_converter,
    v_relative,
    validate_type,
)


def test_intrinsic_insulation_ensemble() -> None:
    """Test the intrinsic insulation ensemble function."""
    assert clo_intrinsic_insulation_ensemble([0.5, 0.5]) == 0.835 + 0.161
    assert clo_intrinsic_insulation_ensemble([1, 1]) == 2 * 0.835 + 0.161
    assert clo_intrinsic_insulation_ensemble(2) == 2 * 0.835 + 0.161
    assert clo_intrinsic_insulation_ensemble([0]) == 0.161


def test_clo_area_factor() -> None:
    """Test the clothing area factor function."""
    assert clo_area_factor(1) == 1.28
    assert np.allclose(clo_area_factor(i_cl=[1, 2]), np.array([1.28, 1.56]))


def test_clo_air_layer_insulation() -> None:
    """Test the clothing insulation air layer function."""
    assert np.isclose(
        clo_insulation_air_layer(vr=1, v_walk=1, i_a_static=0.71),
        0.365,
        atol=0.001,
    )
    assert np.isclose(
        clo_insulation_air_layer(vr=0.2, v_walk=1, i_a_static=0.71),
        0.532,
        atol=0.001,
    )
    assert np.allclose(
        clo_insulation_air_layer(vr=[0.2, 1], v_walk=1, i_a_static=0.71),
        [0.532, 0.365],
        atol=0.001,
    )


def test_clo_total_insulation() -> None:
    """Test the total clothing insulation function."""
    assert np.allclose(
        clo_total_insulation(
            i_t=[1.21, 1.26, 1.56],
            vr=0.15,
            v_walk=0,
            i_a_static=0.5,
            i_cl=[0.61, 0.71, 1.01],
        ),
        [1.21, 1.26, 1.56],
        atol=0.001,
    )

    # compare the normal_clothing results with the figure in the standard
    assert np.allclose(
        clo_total_insulation(
            i_t=[1.21, 1.26, 1.56],
            vr=2,
            v_walk=[1, 0.5, 0.25],
            i_a_static=0.5,
            i_cl=[0.61, 0.71, 1.01],
        ),
        [1.21 * 0.5, 1.26 * 0.565, 1.56 * 0.62],
        atol=0.005,
    )

    # test that the nude function works as expected
    assert np.allclose(
        clo_total_insulation(
            i_t=0,
            vr=0.15,
            v_walk=0,
            i_a_static=[0.71, 0.61, 0.5],
            i_cl=0,
        ),
        [0.71, 0.61, 0.5],
        atol=0.001,
    )

    # compare the nude results with the figure in the standard
    assert np.allclose(
        clo_total_insulation(
            i_t=0,
            vr=[0.5, 2, 3],
            v_walk=0.5,
            i_a_static=[0.71, 0.61, 0.5],
            i_cl=0,
        ),
        [0.71 * 0.7, 0.61 * 0.4, 0.50 * 0.32],
        atol=0.004,
    )

    # test that the low_clothing function works as expected
    assert np.allclose(
        clo_total_insulation(
            i_t=[1.2, 0.6],
            vr=0.15,
            v_walk=0,
            i_a_static=[0.6, 0.6],
            i_cl=[0.6, 0],
        ),
        [1.2, 0.6],
        atol=0.001,
    )

    clo = 0.3
    i_a = 0.7
    np.isclose(
        clo_total_insulation(
            i_t=clo + i_a,
            vr=0.26,
            v_walk=0.06,
            i_a_static=i_a,
            i_cl=clo,
        ),
        0.79,
        atol=0.01,
    )


def test_clo_correction_factor_environment() -> None:
    """Test the clothing correction factor for environment function."""
    assert np.allclose(
        clo_correction_factor_environment(
            vr=0.15,
            v_walk=0,
            i_cl=[0.61, 0.71, 1.01],
        ),
        [1, 1, 1],
        atol=0.001,
    )

    # compare the normal_clothing results with the figure in the standard
    assert np.allclose(
        clo_correction_factor_environment(
            vr=2,
            v_walk=[1, 0.5, 0.25],
            i_cl=[0.61, 0.71, 1.01],
        ),
        [0.503, 0.564, 0.618],
        atol=0.001,
    )

    # test that the nude function works as expected
    assert np.allclose(
        clo_correction_factor_environment(
            vr=0.15,
            v_walk=0,
            i_cl=0,
        ),
        [1],
        atol=0.001,
    )

    # compare the nude results with the figure in the standard
    assert np.allclose(
        clo_correction_factor_environment(
            vr=[0.5, 2, 3],
            v_walk=0.5,
            i_cl=0,
        ),
        [0.698, 0.394, 0.320],
        atol=0.001,
    )

    # test that the low_clothing function works as expected
    assert np.allclose(
        clo_correction_factor_environment(
            vr=0.15,
            v_walk=0,
            i_cl=[0.6, 0],
        ),
        [1, 1],
        atol=0.001,
    )


def test_transpose_sharp_altitude() -> None:
    """Test the transpose_sharp_altitude function."""
    assert transpose_sharp_altitude(sharp=0, altitude=0) == (0, 90)
    assert transpose_sharp_altitude(sharp=0, altitude=20) == (0, 70)
    assert transpose_sharp_altitude(sharp=0, altitude=45) == (0, 45)
    assert transpose_sharp_altitude(sharp=0, altitude=60) == (0, 30)
    assert transpose_sharp_altitude(sharp=90, altitude=0) == (90, 0)
    assert transpose_sharp_altitude(sharp=90, altitude=45) == (45, 0)
    assert transpose_sharp_altitude(sharp=90, altitude=30) == (60, 0)
    assert transpose_sharp_altitude(sharp=135, altitude=60) == (22.208, 20.705)
    assert transpose_sharp_altitude(sharp=120, altitude=75) == (13.064, 7.435)
    assert transpose_sharp_altitude(sharp=150, altitude=30) == (40.893, 48.590)


def test_f_svv() -> None:
    """Test the f_svv function for calculating the clothing insulation factor."""
    assert np.isclose(round(f_svv(30, 10, 3.3), 2), 0.27, atol=1e-09)
    assert np.isclose(round(f_svv(150, 10, 3.3), 2), 0.31, atol=1e-09)
    assert np.isclose(round(f_svv(30, 6, 3.3), 2), 0.20, atol=1e-09)
    assert np.isclose(round(f_svv(150, 6, 3.3), 2), 0.23, atol=1e-09)
    assert np.isclose(round(f_svv(30, 10, 6), 2), 0.17, atol=1e-09)
    assert np.isclose(round(f_svv(150, 10, 6), 2), 0.21, atol=1e-09)
    assert np.isclose(round(f_svv(30, 6, 6), 2), 0.11, atol=1e-09)
    assert np.isclose(round(f_svv(150, 6, 6), 2), 0.14, atol=1e-09)
    assert np.isclose(round(f_svv(6, 9, 3.3), 2), 0.14, atol=1e-09)
    assert np.isclose(round(f_svv(6, 6, 3.3), 2), 0.11, atol=1e-09)
    assert np.isclose(round(f_svv(6, 6, 6), 2), 0.04, atol=1e-09)
    assert np.isclose(round(f_svv(4, 4, 3.3), 2), 0.06, atol=1e-09)
    assert np.isclose(round(f_svv(4, 4, 6), 2), 0.02, atol=1e-09)


def test_running_mean_outdoor_temperature() -> None:
    """Test the running mean outdoor temperature function."""
    assert (running_mean_outdoor_temperature([20, 20], alpha=0.7)) == 20
    assert (running_mean_outdoor_temperature([20, 20], alpha=0.9)) == 20
    assert (running_mean_outdoor_temperature([20, 20, 20, 20], alpha=0.7)) == 20
    assert (running_mean_outdoor_temperature([20, 20, 20, 20], alpha=0.5)) == 20
    assert (
        running_mean_outdoor_temperature(
            [77, 77, 77, 77, 77, 77, 77],
            alpha=0.8,
            units=Units.IP.value,
        )
    ) == 77
    assert (
        running_mean_outdoor_temperature(
            [77, 77, 77, 77, 77, 77, 77],
            alpha=0.8,
            units=Units.IP.value,
        )
    ) == 77


def test_ip_units_converter() -> None:
    """Test the units converter for IP and SI units."""
    assert (units_converter(tdb=77, tr=77, v=3.2, from_units=Units.IP.value)) == [
        25.0,
        25.0,
        0.975312404754648,
    ]
    assert (units_converter(pressure=1, area=1 / 0.09, from_units=Units.IP.value)) == [
        101325,
        1.0322474090590033,
    ]

    expected_result = [25.0, 3.047]
    assert np.allclose(
        units_converter(Units.IP.value, tdb=77, v=10),
        expected_result,
        atol=0.01,
    )

    # Test case 2: Conversion from SI to IP for temperature and velocity
    expected_result = [68, 6.562]
    assert np.allclose(
        units_converter(Units.SI.value, tdb=20, v=2),
        expected_result,
        atol=0.01,
    )

    # Test case 3: Conversion from IP to SI for area and pressure
    expected_result = [9.29, 1489477.5]
    assert np.allclose(
        units_converter(Units.IP.value, area=100, pressure=14.7),
        expected_result,
        atol=0.01,
    )

    # Test case 4: Conversion from SI to IP for area and pressure
    expected_result = [538.199, 1]
    assert np.allclose(
        units_converter(Units.SI.value, area=50, pressure=101325),
        expected_result,
        atol=0.01,
    )


def test_clo_dynamic_ashrae() -> None:
    """Test the dynamic clothing insulation function for ASHRAE standards."""
    assert clo_dynamic_ashrae(clo=1, met=1) == 1
    assert clo_dynamic_ashrae(clo=1, met=0.5) == 1
    assert clo_dynamic_ashrae(clo=2, met=0.5) == 2
    assert np.allclose(clo_dynamic_ashrae(1.0, 1.0), np.array(1))
    assert np.allclose(clo_dynamic_ashrae(1.0, 1.2), np.array(1))
    assert np.allclose(clo_dynamic_ashrae(1.0, 2.0), np.array(0.8))

    # Test invalid standard input
    with pytest.raises(ValueError):
        clo_dynamic_ashrae(1.0, 1.0, model="invalid")


def test_clo_dynamic_iso() -> None:
    """Test the dynamic clothing insulation function for ISO standards."""
    assert np.isclose(clo_dynamic_iso(clo=1, met=1, v=0.2), 0.99, atol=0.01)
    assert np.allclose(
        clo_dynamic_iso(clo=[1, 1.5], met=1, v=0.2),
        [0.99, 1.48],
        atol=0.01,
    )
    assert np.allclose(
        clo_dynamic_iso(clo=[1, 1.5], met=1, v=0.2),
        [0.99, 1.48],
        atol=0.01,
    )
    assert np.allclose(
        clo_dynamic_iso(
            clo=[0.95, 1.07, 0.88, 0.59, 0.83, 0.66, 1.02, 0.71, 1.1, 0.68, 0.3],
            met=[1.71, 1.11, 1.21, 1.77, 1.48, 1.5, 1.33, 1.33, 1.26, 1.47, 1.27],
            v=[0.03, 0.08, 0.04, 0.03, 0.15, 0.15, 0.06, 0.03, 0.25, 0.05, 0.12],
        ),
        [0.85, 1.06, 0.86, 0.52, 0.76, 0.61, 0.97, 0.68, 1.03, 0.63, 0.17],
        atol=0.01,
    )

    # Test invalid standard input
    with pytest.raises(ValueError):
        clo_dynamic_iso(1.0, 1.0, v=0.2, model="invalid")


def test_body_surface_area() -> None:
    """Test the body surface area calculations with various formulas."""
    assert body_surface_area(weight=80, height=1.8) == pytest.approx(1.9917, rel=1e-2)
    assert body_surface_area(70, 1.8, "dubois") == pytest.approx(1.88, rel=1e-2)
    assert body_surface_area(75, 1.75, "takahira") == pytest.approx(1.91, rel=1e-2)
    assert body_surface_area(80, 1.7, "fujimoto") == pytest.approx(1.872, rel=1e-2)
    assert body_surface_area(85, 1.65, "kurazumi") == pytest.approx(1.89, rel=1e-2)
    with pytest.raises(ValueError):
        body_surface_area(70, 1.8, "invalid_formula")


def test_v_relative() -> None:
    """Test the v_relative function for calculating relative air speed."""
    # Test case when met is equal to or lower than 1
    v = 2.0
    met = 1.0
    expected_result = v
    assert np.allclose(v_relative(v, met), expected_result)

    # Test case when met is greater than 1
    v = np.array([1.0, 2.0, 3.0])
    met = 2.0
    expected_result = np.array([1.3, 2.3, 3.3])
    assert np.allclose(v_relative(v, met), expected_result, atol=1e-6)

    # Test case with negative values for v
    v = -1.5
    met = 1.5
    expected_result = -1.5 + 0.3 * 0.5
    assert np.allclose(v_relative(v, met), expected_result, atol=1e-6)


def test_validate_type() -> None:
    """Test the validate_type function for type validation."""
    allowed = (float, int, list, np.ndarray)

    # valid cases
    # native Python types
    validate_type(1, "int_value", allowed)
    validate_type(3.1415, "float_value", allowed)
    validate_type([1, 2, 3], "list_value", allowed)
    validate_type(np.array([1, 2, 3]), "array_value", allowed)

    # np scalars should be converted to native types via .item()
    validate_type(np.float32(40.0), "np_float32", allowed)
    validate_type(np.int32(100), "np_int32", allowed)
    validate_type(np.int64(200), "np_int64", allowed)

    # np array of floats and ints should be allowed
    arr_numeric = np.array([np.float32(1.0), np.int32(2), 3, 3.52])
    validate_type(arr_numeric, "arr_numeric", allowed)

    # empty NumPy array should also pass
    validate_type(np.array([]), "empty_array", allowed)

    # empty list should pass
    validate_type([], "empty_list", allowed)

    # --- Invalid cases ---

    with pytest.raises(TypeError) as exc_info:
        validate_type({"a": 1}, "dict_type", allowed)
    assert "dict_type must be one of the following types:" in str(exc_info.value)

    with pytest.raises(TypeError) as exc_info:
        validate_type("hello", "str_value", allowed)
    assert "str_value must be one of the following types:" in str(exc_info.value)

    with pytest.raises(TypeError) as exc_info:
        validate_type(np.str_("hello"), "np_str", allowed)
    assert "np_str must be one of the following types:" in str(exc_info.value)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import DIInputs
from pythermalcomfort.classes_return import DI
from pythermalcomfort.shared_functions import mapping


def discomfort_index(
    tdb: float | list[float],
    rh: float | list[float],
) -> DI:
    """Calculate the Discomfort Index (DI).

    The index is essentially an
    effective temperature based on air temperature and humidity. The discomfort
    index is usually divided into 6 discomfort categories and it only applies to
    warm environments: [Polydoros2015]_

    * class 1 - DI < 21 °C - No discomfort
    * class 2 - 21 <= DI < 24 °C - Less than 50% feels discomfort
    * class 3 - 24 <= DI < 27 °C - More than 50% feels discomfort
    * class 4 - 27 <= DI < 29 °C - Most of the population feels discomfort
    * class 5 - 29 <= DI < 32 °C - Everyone feels severe stress
    * class 6 - DI >= 32 °C - State of medical emergency

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh : float or list of floats
        Relative humidity, [%].

    Returns
    -------
    DI
        A dataclass containing the Discomfort Index and its classification. See :py:class:`~pythermalcomfort.classes_return.DI` for more details.
        To access the `di` and `discomfort_condition` values, use the respective attributes of the returned `DI` instance, e.g., `result.di`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import discomfort_index

        result = discomfort_index(tdb=25, rh=50)
        print(result.di)  # 22.1
        print(result.discomfort_condition)  # Less than 50% feels discomfort

        result = discomfort_index(tdb=[25, 30], rh=[50, 60])
        print(result.di)  # [22.1, 27.3]
        print(
            result.discomfort_condition
        )  # ['Less than 50% feels discomfort', 'Most of the population feels discomfort']

    """
    # Validate inputs using the DiscomfortIndexInputs class
    DIInputs(
        tdb=tdb,
        rh=rh,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)

    di = tdb - 0.55 * (1 - 0.01 * rh) * (tdb - 14.5)

    di_categories = {
        21: "No discomfort",
        24: "Less than 50% feels discomfort",
        27: "More than 50% feels discomfort",
        29: "Most of the population feels discomfort",
        32: "Everyone feels severe stress",
        99: "State of medical emergency",
    }

    return DI(
        di=np.around(di, 1),
        discomfort_condition=mapping(di, di_categories, right=False),
    )



"""pythermalcomfort: A Python package for thermal comfort calculations.

This package provides comprehensive tools for calculating thermal comfort indices,
heat/cold stress metrics, and thermophysiological responses using multiple models.
"""

__version__ = "3.4.4"



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import WorkCapacityHothapsInputs, WorkIntensity
from pythermalcomfort.classes_return import WorkCapacity


def work_capacity_dunne(
    wbgt: float | list[float],
    work_intensity: str = WorkIntensity.HEAVY.value,
) -> WorkCapacity:
    """Estimate work capacity due to heat based on Dunne et al [Dunne2013]_.

    Estimates the amount of work that will be done at a given WBGT and
    intensity of work as a percent. 100% means work is unaffected by heat. 0%
    means no work is done.

    This is based upon NIOSH safety standards. See:
    Dunne JP, Stouffer RJ, John JG. Reductions in labour capacity from heat
    stress under climate warming. Nature Climate Change. 2013 Jun;3(6):563 6.

    Heavy intensity work is sometimes labelled as 400 W, but this is only
    nominal. Moderate work is assumed to be half as much as heavy intensity, and
    light half as much as moderate.

    Parameters
    ----------
    wbgt : float or list of floats
        Wet bulb globe temperature, [°C].
    work_intensity : str
        Which work intensity to use for the calculation, choice of "heavy",
        "moderate" or "light". Default is "heavy".

        .. note::
            Dunne et al [Dunne2013]_ suggests that heavy intensity work is 350-500 kcal/h, moderate
            is 200-350 kcal/h, and light is less than 100-200 kcal/h.

    Returns
    -------
    WorkCapacity
        A dataclass containing the work capacity. See
        :py:class:`~pythermalcomfort.classes_return.WorkCapacity` for more details. To access the
        `capacity` value, use the `capacity` attribute of the returned `WorkCapacity` instance, e.g.,
        `result.capacity`.

    """
    # validate inputs
    WorkCapacityHothapsInputs(wbgt=wbgt, work_intensity=work_intensity)

    # convert str to enum
    work_intensity = WorkIntensity(work_intensity.lower())
    wbgt = np.array(wbgt)

    capacity = np.clip((100 - (25 * (np.maximum(0, wbgt - 25)) ** (2 / 3))), 0, 100)

    factor_map = {
        WorkIntensity.HEAVY: 1,
        WorkIntensity.MODERATE: 2,
        WorkIntensity.LIGHT: 4,
    }

    capacity = np.clip(capacity * factor_map[work_intensity], 0, 100)

    return WorkCapacity(capacity=capacity)



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import HIInputs
from pythermalcomfort.classes_return import HI
from pythermalcomfort.shared_functions import mapping


def heat_index_rothfusz(
    tdb: float | list[float],
    rh: float | list[float],
    round_output: bool = True,
    limit_inputs: bool = True,
) -> HI:
    """Calculate the Heat Index (HI) in accordance with the Rothfusz (1990) model [Rothfusz1990]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    rh : float or list of floats
        Relative humidity, [%].
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    HI
        A dataclass containing the Heat Index. See :py:class:`~pythermalcomfort.classes_return.HI` for more details.
        To access the `hi` value, use the `hi` attribute of the returned `HI` instance, e.g., `result.hi`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import heat_index_rothfusz

        result = heat_index_rothfusz(tdb=29, rh=50)
        print(result.hi)  # 29.7
        print(result.stress_category)  # "caution"

    """
    # Validate inputs using the HeatIndexInputs class
    HIInputs(
        tdb=tdb,
        rh=rh,
        round_output=round_output,
        limit_inputs=limit_inputs,
    )

    tdb = np.array(tdb)
    rh = np.array(rh)

    hi = -8.784695 + 1.61139411 * tdb + 2.338549 * rh - 0.14611605 * tdb * rh
    hi += -1.2308094 * 10**-2 * tdb**2 - 1.6424828 * 10**-2 * rh**2
    hi += 2.211732 * 10**-3 * tdb**2 * rh + 7.2546 * 10**-4 * tdb * rh**2
    hi += -3.582 * 10**-6 * tdb**2 * rh**2

    # heat index should only be calculated for temperatures above 27 °C
    if limit_inputs:
        tdb_valid = np.where((tdb >= 27.0), tdb, np.nan)
        all_valid = ~(np.isnan(tdb_valid))
        hi_valid = np.where(all_valid, hi, np.nan)
    else:
        hi_valid = hi

    heat_index_categories = {
        27.0: "no risk",
        32.0: "caution",
        41.0: "extreme caution",
        54.0: "danger",
        1000.0: "extreme danger",
    }

    if round_output:
        hi_valid = np.around(hi_valid, 1)

    return HI(hi=hi_valid, stress_category=mapping(hi_valid, heat_index_categories))



import numpy as np
import pytest

from pythermalcomfort.models.ankle_draft import AnkleDraft, ankle_draft
from pythermalcomfort.utilities import Units
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_ankle_draft(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the ankle draft correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.ANKLE_DRAFT.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = ankle_draft(**inputs)

        validate_result(result, outputs, tolerance)


def test_ankle_draft_invalid_input_range() -> None:
    """Test for ankle_draft with invalid input ranges."""
    # Test for ValueError
    with pytest.raises(ValueError):
        ankle_draft(25, 25, 0.3, 50, 1.2, 0.5, 7)
    # Test for ValueError
    with pytest.raises(ValueError):
        ankle_draft(25, 25, 0.3, 50, 1.2, 0.5, [0.1, 0.2, 0.3, 0.4])


def test_ankle_draft_outside_ashrae_range() -> None:
    """Test for ankle_draft with inputs outside the ASHRAE range."""
    r = ankle_draft(50, 25, 0.1, 50, 1.2, 0.5, 0.1)
    assert np.isnan(r.ppd_ad)

    r = ankle_draft([50, 45], 25, 0.1, 50, 1.2, 0.5, 0.1)
    assert np.all(np.isnan(r.ppd_ad))


def test_ankle_draft_list_inputs() -> None:
    """Test for ankle_draft with list inputs."""
    results = ankle_draft(
        tdb=[25, 26, 27],
        tr=[25, 26, 27],
        vr=[0.1, 0.1, 0.1],
        rh=[50, 50, 50],
        met=[1.2, 1.2, 1.2],
        clo=[0.5, 0.5, 0.5],
        v_ankle=[0.1, 0.1, 0.1],
        units=Units.SI.value,
    )
    assert isinstance(results, AnkleDraft)
    assert len(results.ppd_ad) == 3
    assert len(results.acceptability) == 3


def test_ankle_draft_invalid_list_inputs() -> None:
    """Test for TypeError with invalid types in list inputs."""
    with pytest.raises(TypeError):
        ankle_draft(
            tdb=[25, "invalid", 27],
            tr=[25, 26, 27],
            vr=[0.1, 0.1, 0.1],
            rh=[50, 50, 50],
            met=[1.2, 1.2, 1.2],
            clo=[0.5, 0.5, 0.5],
            v_ankle=[0.1, 0.1, 0.1],
            units=Units.SI.value,
        )
